// Code generated by "stringer -type=EventType"; DO NOT EDIT.

package tui

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[Rune-0]
	_ = x[CtrlA-1]
	_ = x[CtrlB-2]
	_ = x[CtrlC-3]
	_ = x[CtrlD-4]
	_ = x[CtrlE-5]
	_ = x[CtrlF-6]
	_ = x[CtrlG-7]
	_ = x[CtrlH-8]
	_ = x[Tab-9]
	_ = x[CtrlJ-10]
	_ = x[CtrlK-11]
	_ = x[CtrlL-12]
	_ = x[Enter-13]
	_ = x[CtrlN-14]
	_ = x[CtrlO-15]
	_ = x[CtrlP-16]
	_ = x[CtrlQ-17]
	_ = x[CtrlR-18]
	_ = x[CtrlS-19]
	_ = x[CtrlT-20]
	_ = x[CtrlU-21]
	_ = x[CtrlV-22]
	_ = x[CtrlW-23]
	_ = x[CtrlX-24]
	_ = x[CtrlY-25]
	_ = x[CtrlZ-26]
	_ = x[Esc-27]
	_ = x[CtrlSpace-28]
	_ = x[CtrlBackSlash-29]
	_ = x[CtrlRightBracket-30]
	_ = x[CtrlCaret-31]
	_ = x[CtrlSlash-32]
	_ = x[ShiftTab-33]
	_ = x[Backspace-34]
	_ = x[Delete-35]
	_ = x[PageUp-36]
	_ = x[PageDown-37]
	_ = x[Up-38]
	_ = x[Down-39]
	_ = x[Left-40]
	_ = x[Right-41]
	_ = x[Home-42]
	_ = x[End-43]
	_ = x[Insert-44]
	_ = x[ShiftUp-45]
	_ = x[ShiftDown-46]
	_ = x[ShiftLeft-47]
	_ = x[ShiftRight-48]
	_ = x[ShiftDelete-49]
	_ = x[ShiftHome-50]
	_ = x[ShiftEnd-51]
	_ = x[ShiftPageUp-52]
	_ = x[ShiftPageDown-53]
	_ = x[F1-54]
	_ = x[F2-55]
	_ = x[F3-56]
	_ = x[F4-57]
	_ = x[F5-58]
	_ = x[F6-59]
	_ = x[F7-60]
	_ = x[F8-61]
	_ = x[F9-62]
	_ = x[F10-63]
	_ = x[F11-64]
	_ = x[F12-65]
	_ = x[AltBackspace-66]
	_ = x[AltUp-67]
	_ = x[AltDown-68]
	_ = x[AltLeft-69]
	_ = x[AltRight-70]
	_ = x[AltDelete-71]
	_ = x[AltHome-72]
	_ = x[AltEnd-73]
	_ = x[AltPageUp-74]
	_ = x[AltPageDown-75]
	_ = x[AltShiftUp-76]
	_ = x[AltShiftDown-77]
	_ = x[AltShiftLeft-78]
	_ = x[AltShiftRight-79]
	_ = x[AltShiftDelete-80]
	_ = x[AltShiftHome-81]
	_ = x[AltShiftEnd-82]
	_ = x[AltShiftPageUp-83]
	_ = x[AltShiftPageDown-84]
	_ = x[CtrlUp-85]
	_ = x[CtrlDown-86]
	_ = x[CtrlLeft-87]
	_ = x[CtrlRight-88]
	_ = x[CtrlHome-89]
	_ = x[CtrlEnd-90]
	_ = x[CtrlBackspace-91]
	_ = x[CtrlDelete-92]
	_ = x[CtrlPageUp-93]
	_ = x[CtrlPageDown-94]
	_ = x[Alt-95]
	_ = x[CtrlAlt-96]
	_ = x[CtrlAltUp-97]
	_ = x[CtrlAltDown-98]
	_ = x[CtrlAltLeft-99]
	_ = x[CtrlAltRight-100]
	_ = x[CtrlAltHome-101]
	_ = x[CtrlAltEnd-102]
	_ = x[CtrlAltBackspace-103]
	_ = x[CtrlAltDelete-104]
	_ = x[CtrlAltPageUp-105]
	_ = x[CtrlAltPageDown-106]
	_ = x[CtrlShiftUp-107]
	_ = x[CtrlShiftDown-108]
	_ = x[CtrlShiftLeft-109]
	_ = x[CtrlShiftRight-110]
	_ = x[CtrlShiftHome-111]
	_ = x[CtrlShiftEnd-112]
	_ = x[CtrlShiftDelete-113]
	_ = x[CtrlShiftPageUp-114]
	_ = x[CtrlShiftPageDown-115]
	_ = x[CtrlAltShiftUp-116]
	_ = x[CtrlAltShiftDown-117]
	_ = x[CtrlAltShiftLeft-118]
	_ = x[CtrlAltShiftRight-119]
	_ = x[CtrlAltShiftHome-120]
	_ = x[CtrlAltShiftEnd-121]
	_ = x[CtrlAltShiftDelete-122]
	_ = x[CtrlAltShiftPageUp-123]
	_ = x[CtrlAltShiftPageDown-124]
	_ = x[Mouse-125]
	_ = x[DoubleClick-126]
	_ = x[LeftClick-127]
	_ = x[RightClick-128]
	_ = x[SLeftClick-129]
	_ = x[SRightClick-130]
	_ = x[ScrollUp-131]
	_ = x[ScrollDown-132]
	_ = x[SScrollUp-133]
	_ = x[SScrollDown-134]
	_ = x[PreviewScrollUp-135]
	_ = x[PreviewScrollDown-136]
	_ = x[Invalid-137]
	_ = x[Fatal-138]
	_ = x[BracketedPasteBegin-139]
	_ = x[BracketedPasteEnd-140]
	_ = x[Resize-141]
	_ = x[Change-142]
	_ = x[BackwardEOF-143]
	_ = x[Start-144]
	_ = x[Load-145]
	_ = x[Focus-146]
	_ = x[One-147]
	_ = x[Zero-148]
	_ = x[Result-149]
	_ = x[Jump-150]
	_ = x[JumpCancel-151]
	_ = x[ClickHeader-152]
	_ = x[ClickFooter-153]
	_ = x[Multi-154]
	_ = x[Every-155]
	_ = x[ResultFinal-156]
}

const _EventType_name = "RuneCtrlACtrlBCtrlCCtrlDCtrlECtrlFCtrlGCtrlHTabCtrlJCtrlKCtrlLEnterCtrlNCtrlOCtrlPCtrlQCtrlRCtrlSCtrlTCtrlUCtrlVCtrlWCtrlXCtrlYCtrlZEscCtrlSpaceCtrlBackSlashCtrlRightBracketCtrlCaretCtrlSlashShiftTabBackspaceDeletePageUpPageDownUpDownLeftRightHomeEndInsertShiftUpShiftDownShiftLeftShiftRightShiftDeleteShiftHomeShiftEndShiftPageUpShiftPageDownF1F2F3F4F5F6F7F8F9F10F11F12AltBackspaceAltUpAltDownAltLeftAltRightAltDeleteAltHomeAltEndAltPageUpAltPageDownAltShiftUpAltShiftDownAltShiftLeftAltShiftRightAltShiftDeleteAltShiftHomeAltShiftEndAltShiftPageUpAltShiftPageDownCtrlUpCtrlDownCtrlLeftCtrlRightCtrlHomeCtrlEndCtrlBackspaceCtrlDeleteCtrlPageUpCtrlPageDownAltCtrlAltCtrlAltUpCtrlAltDownCtrlAltLeftCtrlAltRightCtrlAltHomeCtrlAltEndCtrlAltBackspaceCtrlAltDeleteCtrlAltPageUpCtrlAltPageDownCtrlShiftUpCtrlShiftDownCtrlShiftLeftCtrlShiftRightCtrlShiftHomeCtrlShiftEndCtrlShiftDeleteCtrlShiftPageUpCtrlShiftPageDownCtrlAltShiftUpCtrlAltShiftDownCtrlAltShiftLeftCtrlAltShiftRightCtrlAltShiftHomeCtrlAltShiftEndCtrlAltShiftDeleteCtrlAltShiftPageUpCtrlAltShiftPageDownMouseDoubleClickLeftClickRightClickSLeftClickSRightClickScrollUpScrollDownSScrollUpSScrollDownPreviewScrollUpPreviewScrollDownInvalidFatalBracketedPasteBeginBracketedPasteEndResizeChangeBackwardEOFStartLoadFocusOneZeroResultJumpJumpCancelClickHeaderClickFooterMultiEveryResultFinal"

var _EventType_index = [...]uint16{0, 4, 9, 14, 19, 24, 29, 34, 39, 44, 47, 52, 57, 62, 67, 72, 77, 82, 87, 92, 97, 102, 107, 112, 117, 122, 127, 132, 135, 144, 157, 173, 182, 191, 199, 208, 214, 220, 228, 230, 234, 238, 243, 247, 250, 256, 263, 272, 281, 291, 302, 311, 319, 330, 343, 345, 347, 349, 351, 353, 355, 357, 359, 361, 364, 367, 370, 382, 387, 394, 401, 409, 418, 425, 431, 440, 451, 461, 473, 485, 498, 512, 524, 535, 549, 565, 571, 579, 587, 596, 604, 611, 624, 634, 644, 656, 659, 666, 675, 686, 697, 709, 720, 730, 746, 759, 772, 787, 798, 811, 824, 838, 851, 863, 878, 893, 910, 924, 940, 956, 973, 989, 1004, 1022, 1040, 1060, 1065, 1076, 1085, 1095, 1105, 1116, 1124, 1134, 1143, 1154, 1169, 1186, 1193, 1198, 1217, 1234, 1240, 1246, 1257, 1262, 1266, 1271, 1274, 1278, 1284, 1288, 1298, 1309, 1320, 1325, 1330, 1341}

func (i EventType) String() string {
	if i < 0 || i >= EventType(len(_EventType_index)-1) {
		return "EventType(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _EventType_name[_EventType_index[i]:_EventType_index[i+1]]
}
