// Code generated by "stringer -type=actionType"; DO NOT EDIT.

package fzf

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[actIgnore-0]
	_ = x[actStart-1]
	_ = x[actClick-2]
	_ = x[actInvalid-3]
	_ = x[actBracketedPasteBegin-4]
	_ = x[actBracketedPasteEnd-5]
	_ = x[actChar-6]
	_ = x[actMouse-7]
	_ = x[actBeginningOfLine-8]
	_ = x[actAbort-9]
	_ = x[actAccept-10]
	_ = x[actAcceptNonEmpty-11]
	_ = x[actAcceptOrPrintQuery-12]
	_ = x[actBackwardChar-13]
	_ = x[actBackwardDeleteChar-14]
	_ = x[actBackwardDeleteCharEof-15]
	_ = x[actBackwardWord-16]
	_ = x[actBackwardSubWord-17]
	_ = x[actCancel-18]
	_ = x[actChangeBorderLabel-19]
	_ = x[actChangeGhost-20]
	_ = x[actChangeHeader-21]
	_ = x[actChangeHeaderLines-22]
	_ = x[actChangeFooter-23]
	_ = x[actChangeHeaderLabel-24]
	_ = x[actChangeFooterLabel-25]
	_ = x[actChangeInputLabel-26]
	_ = x[actChangeListLabel-27]
	_ = x[actChangeMulti-28]
	_ = x[actChangeNth-29]
	_ = x[actChangeWithNth-30]
	_ = x[actChangePointer-31]
	_ = x[actChangePreview-32]
	_ = x[actChangePreviewLabel-33]
	_ = x[actChangePreviewWindow-34]
	_ = x[actChangePrompt-35]
	_ = x[actChangeQuery-36]
	_ = x[actClearScreen-37]
	_ = x[actClearQuery-38]
	_ = x[actClearSelection-39]
	_ = x[actClose-40]
	_ = x[actDeleteChar-41]
	_ = x[actDeleteCharEof-42]
	_ = x[actEndOfLine-43]
	_ = x[actFatal-44]
	_ = x[actForwardChar-45]
	_ = x[actForwardWord-46]
	_ = x[actForwardSubWord-47]
	_ = x[actKillLine-48]
	_ = x[actKillWord-49]
	_ = x[actKillSubWord-50]
	_ = x[actUnixLineDiscard-51]
	_ = x[actUnixWordRubout-52]
	_ = x[actYank-53]
	_ = x[actBackwardKillWord-54]
	_ = x[actBackwardKillSubWord-55]
	_ = x[actSelectAll-56]
	_ = x[actDeselectAll-57]
	_ = x[actToggle-58]
	_ = x[actToggleSearch-59]
	_ = x[actToggleAll-60]
	_ = x[actToggleDown-61]
	_ = x[actToggleUp-62]
	_ = x[actToggleIn-63]
	_ = x[actToggleOut-64]
	_ = x[actToggleTrack-65]
	_ = x[actToggleTrackCurrent-66]
	_ = x[actToggleHeader-67]
	_ = x[actToggleWrap-68]
	_ = x[actToggleWrapWord-69]
	_ = x[actToggleMultiLine-70]
	_ = x[actToggleHscroll-71]
	_ = x[actToggleRaw-72]
	_ = x[actEnableRaw-73]
	_ = x[actDisableRaw-74]
	_ = x[actTrackCurrent-75]
	_ = x[actToggleInput-76]
	_ = x[actHideInput-77]
	_ = x[actShowInput-78]
	_ = x[actUntrackCurrent-79]
	_ = x[actDown-80]
	_ = x[actDownMatch-81]
	_ = x[actUp-82]
	_ = x[actUpMatch-83]
	_ = x[actPageUp-84]
	_ = x[actPageDown-85]
	_ = x[actPosition-86]
	_ = x[actHalfPageUp-87]
	_ = x[actHalfPageDown-88]
	_ = x[actOffsetUp-89]
	_ = x[actOffsetDown-90]
	_ = x[actOffsetMiddle-91]
	_ = x[actJump-92]
	_ = x[actJumpAccept-93]
	_ = x[actPrintQuery-94]
	_ = x[actRefreshPreview-95]
	_ = x[actReplaceQuery-96]
	_ = x[actToggleSort-97]
	_ = x[actShowPreview-98]
	_ = x[actHidePreview-99]
	_ = x[actTogglePreview-100]
	_ = x[actTogglePreviewWrap-101]
	_ = x[actTogglePreviewWrapWord-102]
	_ = x[actTransform-103]
	_ = x[actTransformBorderLabel-104]
	_ = x[actTransformGhost-105]
	_ = x[actTransformHeader-106]
	_ = x[actTransformHeaderLines-107]
	_ = x[actTransformFooter-108]
	_ = x[actTransformHeaderLabel-109]
	_ = x[actTransformFooterLabel-110]
	_ = x[actTransformInputLabel-111]
	_ = x[actTransformListLabel-112]
	_ = x[actTransformNth-113]
	_ = x[actTransformWithNth-114]
	_ = x[actTransformPointer-115]
	_ = x[actTransformPreviewLabel-116]
	_ = x[actTransformPrompt-117]
	_ = x[actTransformQuery-118]
	_ = x[actTransformSearch-119]
	_ = x[actTrigger-120]
	_ = x[actBgTransform-121]
	_ = x[actBgTransformBorderLabel-122]
	_ = x[actBgTransformGhost-123]
	_ = x[actBgTransformHeader-124]
	_ = x[actBgTransformHeaderLines-125]
	_ = x[actBgTransformFooter-126]
	_ = x[actBgTransformHeaderLabel-127]
	_ = x[actBgTransformFooterLabel-128]
	_ = x[actBgTransformInputLabel-129]
	_ = x[actBgTransformListLabel-130]
	_ = x[actBgTransformNth-131]
	_ = x[actBgTransformWithNth-132]
	_ = x[actBgTransformPointer-133]
	_ = x[actBgTransformPreviewLabel-134]
	_ = x[actBgTransformPrompt-135]
	_ = x[actBgTransformQuery-136]
	_ = x[actBgTransformSearch-137]
	_ = x[actBgCancel-138]
	_ = x[actSearch-139]
	_ = x[actPreview-140]
	_ = x[actPreviewTop-141]
	_ = x[actPreviewBottom-142]
	_ = x[actPreviewUp-143]
	_ = x[actPreviewDown-144]
	_ = x[actPreviewPageUp-145]
	_ = x[actPreviewPageDown-146]
	_ = x[actPreviewHalfPageUp-147]
	_ = x[actPreviewHalfPageDown-148]
	_ = x[actPrevHistory-149]
	_ = x[actPrevSelected-150]
	_ = x[actPrint-151]
	_ = x[actPut-152]
	_ = x[actNextHistory-153]
	_ = x[actNextSelected-154]
	_ = x[actExecute-155]
	_ = x[actExecuteSilent-156]
	_ = x[actExecuteMulti-157]
	_ = x[actSigStop-158]
	_ = x[actBest-159]
	_ = x[actFirst-160]
	_ = x[actLast-161]
	_ = x[actReload-162]
	_ = x[actReloadSync-163]
	_ = x[actDisableSearch-164]
	_ = x[actEnableSearch-165]
	_ = x[actSelect-166]
	_ = x[actDeselect-167]
	_ = x[actUnbind-168]
	_ = x[actRebind-169]
	_ = x[actToggleBind-170]
	_ = x[actBecome-171]
	_ = x[actShowHeader-172]
	_ = x[actHideHeader-173]
	_ = x[actBell-174]
	_ = x[actExclude-175]
	_ = x[actExcludeMulti-176]
	_ = x[actAsync-177]
	_ = x[actWait-178]
}

const _actionType_name = "actIgnoreactStartactClickactInvalidactBracketedPasteBeginactBracketedPasteEndactCharactMouseactBeginningOfLineactAbortactAcceptactAcceptNonEmptyactAcceptOrPrintQueryactBackwardCharactBackwardDeleteCharactBackwardDeleteCharEofactBackwardWordactBackwardSubWordactCancelactChangeBorderLabelactChangeGhostactChangeHeaderactChangeHeaderLinesactChangeFooteractChangeHeaderLabelactChangeFooterLabelactChangeInputLabelactChangeListLabelactChangeMultiactChangeNthactChangeWithNthactChangePointeractChangePreviewactChangePreviewLabelactChangePreviewWindowactChangePromptactChangeQueryactClearScreenactClearQueryactClearSelectionactCloseactDeleteCharactDeleteCharEofactEndOfLineactFatalactForwardCharactForwardWordactForwardSubWordactKillLineactKillWordactKillSubWordactUnixLineDiscardactUnixWordRuboutactYankactBackwardKillWordactBackwardKillSubWordactSelectAllactDeselectAllactToggleactToggleSearchactToggleAllactToggleDownactToggleUpactToggleInactToggleOutactToggleTrackactToggleTrackCurrentactToggleHeaderactToggleWrapactToggleWrapWordactToggleMultiLineactToggleHscrollactToggleRawactEnableRawactDisableRawactTrackCurrentactToggleInputactHideInputactShowInputactUntrackCurrentactDownactDownMatchactUpactUpMatchactPageUpactPageDownactPositionactHalfPageUpactHalfPageDownactOffsetUpactOffsetDownactOffsetMiddleactJumpactJumpAcceptactPrintQueryactRefreshPreviewactReplaceQueryactToggleSortactShowPreviewactHidePreviewactTogglePreviewactTogglePreviewWrapactTogglePreviewWrapWordactTransformactTransformBorderLabelactTransformGhostactTransformHeaderactTransformHeaderLinesactTransformFooteractTransformHeaderLabelactTransformFooterLabelactTransformInputLabelactTransformListLabelactTransformNthactTransformWithNthactTransformPointeractTransformPreviewLabelactTransformPromptactTransformQueryactTransformSearchactTriggeractBgTransformactBgTransformBorderLabelactBgTransformGhostactBgTransformHeaderactBgTransformHeaderLinesactBgTransformFooteractBgTransformHeaderLabelactBgTransformFooterLabelactBgTransformInputLabelactBgTransformListLabelactBgTransformNthactBgTransformWithNthactBgTransformPointeractBgTransformPreviewLabelactBgTransformPromptactBgTransformQueryactBgTransformSearchactBgCancelactSearchactPreviewactPreviewTopactPreviewBottomactPreviewUpactPreviewDownactPreviewPageUpactPreviewPageDownactPreviewHalfPageUpactPreviewHalfPageDownactPrevHistoryactPrevSelectedactPrintactPutactNextHistoryactNextSelectedactExecuteactExecuteSilentactExecuteMultiactSigStopactBestactFirstactLastactReloadactReloadSyncactDisableSearchactEnableSearchactSelectactDeselectactUnbindactRebindactToggleBindactBecomeactShowHeaderactHideHeaderactBellactExcludeactExcludeMultiactAsyncactWait"

var _actionType_index = [...]uint16{0, 9, 17, 25, 35, 57, 77, 84, 92, 110, 118, 127, 144, 165, 180, 201, 225, 240, 258, 267, 287, 301, 316, 336, 351, 371, 391, 410, 428, 442, 454, 470, 486, 502, 523, 545, 560, 574, 588, 601, 618, 626, 639, 655, 667, 675, 689, 703, 720, 731, 742, 756, 774, 791, 798, 817, 839, 851, 865, 874, 889, 901, 914, 925, 936, 948, 962, 983, 998, 1011, 1028, 1046, 1062, 1074, 1086, 1099, 1114, 1128, 1140, 1152, 1169, 1176, 1188, 1193, 1203, 1212, 1223, 1234, 1247, 1262, 1273, 1286, 1301, 1308, 1321, 1334, 1351, 1366, 1379, 1393, 1407, 1423, 1443, 1467, 1479, 1502, 1519, 1537, 1560, 1578, 1601, 1624, 1646, 1667, 1682, 1701, 1720, 1744, 1762, 1779, 1797, 1807, 1821, 1846, 1865, 1885, 1910, 1930, 1955, 1980, 2004, 2027, 2044, 2065, 2086, 2112, 2132, 2151, 2171, 2182, 2191, 2201, 2214, 2230, 2242, 2256, 2272, 2290, 2310, 2332, 2346, 2361, 2369, 2375, 2389, 2404, 2414, 2430, 2445, 2455, 2462, 2470, 2477, 2486, 2499, 2515, 2530, 2539, 2550, 2559, 2568, 2581, 2590, 2603, 2616, 2623, 2633, 2648, 2656, 2663}

func (i actionType) String() string {
	if i < 0 || i >= actionType(len(_actionType_index)-1) {
		return "actionType(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _actionType_name[_actionType_index[i]:_actionType_index[i+1]]
}
